# -*- coding: utf-8 -*-

# Initialization on Editha: source /afs/mpi-magdeburg.mpg.de/data/csc/software/64bit/fenics-1.5.0-ubuntu1204/share/dolfin/dolfin.conf
# Modify nz,ny,nz in FlowProblem2Optimize.run,
# K0, x0, c in main

from dolfin import *
from block import *
from block.algebraic.trilinos import ML 
from block.dolfin_util import *
from block.iterative import *
import scipy.io as sio
class EssentialBC(Expression):
    def eval(self, v, x):
        if x[0] < DOLFIN_EPS:
            y = x[1]
            v[0] = y*(1-y)
	    v[1] = 0
            v[2] = 0
        else:
            v[0] = 0; v[1] = 0;v[2] = 0

    def value_shape(self):
  	      return (3,)

class K(Expression):
    def __init__(self, K0, x0, c):
        self.K0, self.x0, self.c = K0, x0, c

    def eval(self, v, x):
        x0, K0, c = self.x0, self.K0, self.c
	if (x[0] - x0)**2 + (x[1] - 0.5)**2+ (x[2] - 2.0)**2 <= c**2:
            v[0] = K0                        
        else:
            v[0] = 1.0

def dirichlet_boundary(x):
    return bool(x[0] < DOLFIN_EPS or x[1] < DOLFIN_EPS or x[2] < DOLFIN_EPS or\
                x[0] > 1.0 - DOLFIN_EPS or x[1] > 1.0 - DOLFIN_EPS)

def dirichlet_boundary_for_p(x):
    return bool(x[0] < DOLFIN_EPS)

def outflow_boundary(x):
    return bool(x[0] > 1.0 - DOLFIN_EPS)

class FlowProblem2Optimize:
    def __init__(self, K0, x0, c, plot):
        self.K0, self.x0, self.c, self.plot = K0, x0, c, plot

    def run(self):
        K0, x0, c = self.K0, self.x0, self.c
        
	nx= 16 # 20
	ny= 16 # 20
	nz= 32 # 42
	mesh = BoxMesh(0.0,0.0,0.0, 1.0, 1.0, 5.0, nx, ny, nz)

	# Define function spaces
	V = VectorFunctionSpace(mesh, "CG", 2)
	Q = FunctionSpace(mesh, "CG", 1)
	k = K(K0, x0, c)


	u_inflow = EssentialBC()
        bc1 = DirichletBC(V, u_inflow, dirichlet_boundary)
	zero = Constant(0)
 	bcp = DirichletBC(Q, zero, dirichlet_boundary_for_p)
	zerov = Constant((0,0,0))
	bc2 = DirichletBC(V, zerov, dirichlet_boundary)
	bc_zero = [bc2,None]
	# Define variational problem and assemble matrices
	v, u = TestFunction(V), TrialFunction(V)
	q, p = TestFunction(Q), TrialFunction(Q)

	f = Constant((0, 0,0))

	# Two functions for the uncertain right hand side
	f0 = Constant((2.0, 2.0, 2.0))
	f1 = Constant((0.1, 2.0, 2.0))
	fdummy = Constant((0.0))

	a11 = inner(grad(u), grad(v))*dx + k*inner(u, v)*dx 
	a11lp = inner(grad(u), grad(v))*dx # for stochastic only
	a12 = div(v)*p*dx
	a21 = div(u)*q*dx
	
	m11 = inner(u,v)*dx
	L1  = inner(v, f)*dx
	F0  = inner(v, f0)*dx
	F1  = inner(v, f1)*dx
	Fdummy  = inner(q, fdummy)*dx

	kp = inner(grad(p), grad(q))*dx # Pressure Laplacian
        mp = inner(p, q)*dx # Pressure Mass matrix
        mpk = k*inner(p, q)*dx # Pressure Mass with k coefficient
        kpk = inner(grad(p), grad(q))*dx+k*inner(p, q)*dx # Pressure Laplacian with k coefficient
	# Assemble matrices for the preconditioner
	M,dummy1 = assemble_system(m11,L1,bcs=bc1) # Velocity mass
	Kp,dummy1 = assemble_system(kp,Fdummy,bcs=bcp) # Velocity mass
	Kpk,dummy1 = assemble_system(kpk,Fdummy,bcs=bcp)
	# Kpk = assemble(kpk)              
        Mp = assemble(mp)
        Mpk = assemble(mpk)


	# Block assemble will take care about BC
	bcs = [bc2,None,bc2,bc1,None]
        AA = block_assemble(          [[a11, a12,   0,   m11, 0],
					         [a21,   0,   0,   0,   0],
                                                 [m11,   0,  m11,  a11lp,   0],   
                                                 [0,     0,  m11,  a11, a12], 
                                                 [0,     0,  0,    a21,  0] ]) # , bcs=bcs
                                              
        bb = block_assemble([0,0,0,F0, 0])

	block_bc(bcs, True).apply(AA).apply(bb)

        # Extract components
        [[A,B,_,M,_],
         [C,_,_,_,_],
         [_,_,M2,A2,_],
         [_,_,_,_,_],
         [_,_,_,_,_]] = AA

	[bb1,bb2,bb3,bb4,bb5] = bb
	
	# Dump to files
	mfile = File ("A.mtx")
	mfile << A
	mfile = File ("A2.mtx")
	mfile << A2	
	mfile = File ("C.mtx")
	mfile << C
	mfile = File ("B.mtx")
	mfile << B
	mfile = File ("M.mtx")
	mfile << M	
	mfile = File ("M2.mtx")
	mfile << M2	


        rfile = File ("rhs1.mtx")
	rfile << bb1
	rfile = File ("rhs2.mtx")
	rfile << bb2
        rfile = File ("rhs3.mtx")
	rfile << bb3
	rfile = File ("rhs4.mtx")
	rfile << bb4
        rfile = File ("rhs5.mtx")
	rfile << bb5
	
        mfile = File ("Kp.mtx")
	mfile << Kp
        mfile = File ("Kpk.mtx")
	mfile << Kpk
	mfile = File ("Mp.mtx")
	mfile << Mp
        mfile = File ("Mpk.mtx")
	mfile << Mpk	
	
	
	# If we need to plot pictures, rerun this
	# Convert matlab vectors to VTK	
	mat_file = sio.loadmat('3dsol_state_mean_last.mat')
	u_vec = mat_file['xv'] 
	u_vec=u_vec.transpose()
	print type(u_vec), u_vec.shape 	
	u_vec = u_vec[0]
	print type(u_vec), u_vec.shape 	
	umatlab = Function(V)
	umatlab.vector()[:] = u_vec
	#plot(Function(V, umatlab))
	file = File('3d_state_mean_last.pvd')
	file << Function(V,umatlab)	
	
	mat_file = sio.loadmat('3dsol_state_p_mean_last.mat')
	u_vec = mat_file['xv'] 
	u_vec=u_vec.transpose()
	print type(u_vec), u_vec.shape 	
	u_vec = u_vec[0]
	print type(u_vec), u_vec.shape 	
	umatlab = Function(Q)
	umatlab.vector()[:] = u_vec
	#plot(Function(V, umatlab))
	file = File('3d_state_p_mean_last.pvd')
	file << Function(Q,umatlab)		

	# Convert matlab vectors to VTK	
	mat_file = sio.loadmat('3dsol_con_mean_last.mat')
	u_vec = mat_file['xv'] 
	u_vec=u_vec.transpose()
	print type(u_vec), u_vec.shape 	
	u_vec = u_vec[0]
	print type(u_vec), u_vec.shape 	
	umatlab = Function(V)
	umatlab.vector()[:] = u_vec
	#plot(Function(V, umatlab))
	file = File('3d_con_mean_last.pvd')
	file << Function(V,umatlab)	

	# Convert matlab vectors to VTK	
	mat_file = sio.loadmat('3dsol_lagr_mean_last.mat')
	u_vec = mat_file['xv'] 
	u_vec=u_vec.transpose()
	print type(u_vec), u_vec.shape 	
	u_vec = u_vec[0]
	print type(u_vec), u_vec.shape 	
	umatlab = Function(V)
	umatlab.vector()[:] = u_vec
	#plot(Function(V, umatlab))
	file = File('3d_lagr_mean_last.pvd')
	file << Function(V,umatlab)	
	
	
	# Convert matlab vectors to VTK	
	mat_file = sio.loadmat('3dsol_state_std_last.mat')
	u_vec = mat_file['xv'] 
	u_vec=u_vec.transpose()
	print type(u_vec), u_vec.shape 	
	u_vec = u_vec[0]
	print type(u_vec), u_vec.shape 	
	umatlab = Function(V)
	umatlab.vector()[:] = u_vec
	#plot(Function(V, umatlab))
	file = File('3d_state_std_last.pvd')
	file << Function(V,umatlab)	
	
	mat_file = sio.loadmat('3dsol_state_p_std_last.mat')
	u_vec = mat_file['xv'] 
	u_vec=u_vec.transpose()
	print type(u_vec), u_vec.shape 	
	u_vec = u_vec[0]
	print type(u_vec), u_vec.shape 	
	umatlab = Function(Q)
	umatlab.vector()[:] = u_vec
	#plot(Function(V, umatlab))
	file = File('3d_state_p_std_last.pvd')
	file << Function(Q,umatlab)		

	# Convert matlab vectors to VTK	
	mat_file = sio.loadmat('3dsol_con_std_last.mat')
	u_vec = mat_file['xv'] 
	u_vec=u_vec.transpose()
	print type(u_vec), u_vec.shape 	
	u_vec = u_vec[0]
	print type(u_vec), u_vec.shape 	
	umatlab = Function(V)
	umatlab.vector()[:] = u_vec
	#plot(Function(V, umatlab))
	file = File('3d_con_std_last.pvd')
	file << Function(V,umatlab)	

	# No idea what is it for
	goal1 = 1
	goal2 = 2
	goal = goal1+goal2
        key_variables = dict(K0=K0, x0=x0, c=c, goal1=goal1,
                             goal2=goal2, goal=goal)
        print key_variables
        return goal1, goal2


if __name__ == "__main__":
    from sys import argv

    # a test with the porous media placed close to the middle of the domain
    K0=1e4; x0 = 0.5; c=0.1
    
    p = FlowProblem2Optimize(K0, x0, c, True)

    p.run()

    interactive()



