% The runfile for a single 3D experiment. Also, to compare two
% preconditioners
% Use gen_SBmat_3d.py to generate matrices
% The local stokes system are solved with HSL AMG

% The hierarchy of procedures is as follows:
% this_routine->amen_block_solve->locsolve
% These global variables are needed to communicate with locsolve directly.
global p_state
global p_lagr
global spat_iters

p_state = [];
p_lagr = [];
spat_iters = 0;

% Time dimension
N = 8;
T = 1;
nt = 2^N;
dt = T/nt; 
% Regularization
beta = 1e-6;
alpha = 1;
% Tolerance 
tol = 1e-4;
% Stochastics
nu1 = 1e-2;
 

%%%%%%%%%%%%% End of parameter section %%%%%%%%%%%%%%%

% Time matrices
It = speye(nt);
Ct = spdiags(-ones(nt,1), -1, nt, nt);

% Stochastic matrices
load G0_G1.mat
G0 = sparse(diag(Epsi2));
G1=  PCE;
G1(abs(G1)<1e-15)=0;
G1 = sparse(G1);


% Spatial matrices from Fenics
A=mmread('A.mtx'); % Stokes+Brinkmann
A2=mmread('A2.mtx'); % Stokes only, without boundary nodes -- for stochastic term
B=mmread('C.mtx'); % Funny

Mv=mmread('M.mtx'); % mass without boundary nodes
Mv2=mmread('M2.mtx'); % with boundary nodes

% Pressure matrices
Mp=mmread('Mp.mtx');
Kp=mmread('Kp.mtx');
Kpk=mmread('Kpk.mtx'); % inner(grad(p),grad(q))+k*inner(p,q)

% RHS
r4=mmread('rhs4.mtx');
r5=mmread('rhs5.mtx');

% Extract the sizes
nv=size(A,1);
np=size(B,1);
P = size(G0,1);

% Solve the forward deterministic problem
% Stokes Matrix
AA = [A B';B sparse(np,np)];
rhs = [r4;r5];

% Velocity preconditioner
global avoid_hsl_prec
if (~isempty(avoid_hsl_prec))&&(~avoid_hsl_prec)
    control = hsl_mi20_control;
    info = hsl_mi20_setup(A, control);
end;

% Prepare the desired state-- the Stokes solution
[x,iter2,resvec2] = fgmres(AA,rhs,1e-9,'restart', 100, 'P', @(x,tol)schurprec_forw(x,A,B',Kpk,Mp,Kp));

if (~isempty(avoid_hsl_prec))&&(~avoid_hsl_prec)
    hsl_mi20_finalize;
end;

xst = x;


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Now, the inverse problem

% To augment Mv, we need its norm
[~,nrmMv] = eigs(Mv,1,'LM');

% More time and stoch matrices
D=speye(nt);
% D(1,1)=1/2; The first time point is dt, not 0! Hence the quadrature weight 1
D(nt,nt)=1/2; % The last is T, ok.
G0 = sparse(G0);
G1 = sparse(G1);

T1 = diag(G0); 
T1(1)= 0;
T1 = alpha*sparse(diag(T1));
G_alpha = G0 + T1;

% The block matrix will have the 3x3 structure:
% M1  0  A'
% 0   M2  N'
% A   N  0  
% Blocks will be stored in the {d,R} format

% The observ. mass
A11 = {dt*Mv; G_alpha; D};

% The control mass
A22 = {dt*beta*Mv2; G0; D};

% The N-matrix
A32 = {dt*Mv; G0; It};

% Now the system matrices
A31 = {dt*(Mv/dt+A), dt*(nu1*A2), dt*(Mv/dt);
       G0,          G1,     G0;
       It,          It,     Ct};   
A13 = {dt*(Mv/dt+A'),dt*(nu1*A2), dt*(Mv/dt);
       G0,          G1,     G0;
       It,          It,     Ct'};   
   
% Accemble everything
Ac = {A11, [], A13;
      [],  A22, A32;
      A31, A32, []};

estoch = tt_unit(P,1,1);
et = tt_ones(nt,1);   
  
% The right-hand side
b1_space = dt*Mv*xst(1:nv);
b1 = mtkron(tt_tensor(b1_space), tt_tensor(G0*full(estoch)), tt_tensor(full(diag(D))));
b2 = tt_zeros([nv; P; nt]);
% The last component with the stationary RHS
b3 = mtkron(tt_tensor(dt*r4), estoch, et);

% Assemble to the cell
bc = {b1;b2;b3};

% Pressure parts of the preconditioner
P_stokes = struct;
P_stokes.Mp = dt*Mp;
P_stokes.Kp = dt*Kp;
P_stokes.Fp = {Mp+dt*Kpk, dt*nu1*Kp, Mp};

% Solve everything
locsolve_prec = 2; % don't experiment with the prec here
x0 = [tt_zeros([nv,P,nt]);tt_zeros([nv,P,nt]);tt_zeros([nv,P,nt])];
auxmat = {
    {[dt*B' sparse(nv,np)];G0;It}; 
    []; 
    {[sparse(nv,np) dt*B'];G0;It}
    }; % B' matrices excluded with the pressures

tic;
sol = amen_block_solve(Ac,bc,tol, 'kickrank', 2, 'resid_damp', 1e1, 'x0', x0, 'nswp', 3, 'exitdir', -1, 'auxmat', auxmat, ...
    'solfun', @(i, XAX1,Ai,XAX2, XY1,yi,XY2, tol, sol_prev, XBX1, Bi, XBX2)locsolve_iter(i, XAX1,Ai,XAX2, XY1,yi,XY2, tol, sol_prev, XBX1,Bi,XBX2, dt*r5, P_stokes));
ttimes = toc;
fprintf('Solution time: %g\n', ttimes);

sol1 = [1 0 0]*sol;
sol2 = [0 1 0]*sol;
sol3 = [0 0 1]*sol;
% Postprocessing -- incorporate the pressure
sol11 = zeros(1, nv+np, size(p_state,2));
sol11(1,1:nv,:)=sol1{1};
sol11(1,nv+1:nv+np,:)=p_state;
sol1{1} = sol11;

sol31 = zeros(1, nv+np, size(p_lagr,2));
sol31(1,1:nv,:)=sol3{1};
sol31(1,nv+1:nv+np,:)=p_lagr;
sol3{1} = sol31;


% Measure the residual
A11c = {A11(1); A11(2); A11(3)};
A22c = {A22(1); A22(2); A22(3)};
A32c = {A32(1); A32(2); A32(3)};
A31c = cell(3,1);
A13c = cell(3,1);
for j=1:3
    A31c{j} = A31(j,:);
    A13c{j} = A13(j,:);
end;
% Expand the PDE to the stokes form
A31c{1}{1} = [A31{1,1}, dt*B'; dt*B, sparse(np,np)];
A31c{1}{2} = blkdiag(A31{1,2}, sparse(np,np));
A31c{1}{3} = blkdiag(A31{1,3}, sparse(np,np));

% The same for adjoint matrix
A13c{1}{1} = [A13{1,1}, dt*B'; dt*B, sparse(np,np)];
A13c{1}{2} = blkdiag(A13{1,2}, sparse(np,np));
A13c{1}{3} = blkdiag(A13{1,3}, sparse(np,np));

% Expand M1 to the stokes form
A11c{1}{1} = blkdiag(A11{1,1}, sparse(np,np));

% N
A32c{1}{1} = [A32{1,1}; sparse(np,nv)];
A23c = A32c;
A23c{1}{1} = A32c{1}{1}';

bc{1}{1} = [bc{1}{1}, zeros(1,np)];
bc{3}{1} = [bc{3}{1}, dt*r5.'];

resid1 = norm(amen_mv(A11c,sol1,tol,'verb',0)+amen_mv(A13c,sol3,tol,'verb',0)-bc{1})/norm(bc{1})
resid2 = norm(amen_mv(A22c,sol2,tol,'verb',0)+amen_mv(A23c,sol3,tol,'verb',0))/norm(amen_mv(A23c,sol3,tol,'verb',0))
resid3 = norm(amen_mv(A31c,sol1,tol,'verb',0)+amen_mv(A32c,sol2,tol,'verb',0)-bc{3})/norm(bc{3})

% Mean observation error 
xst_ten = mtkron(tt_tensor(xst(1:nv)), estoch, et);
err_obs = norm(sol1(1:nv,:,:)-xst_ten)/norm(xst_ten) % over all DOFs
err_obs_T = norm(full(sol1(1:nv, 1, nt))-xst(1:nv))/norm(xst(1:nv)) % at the final time

% Moments of the solution
sol1_std = round(sol1(1:nv, :, :), tol);
sol1_std = dot(sol1_std, amen_mv({{dt*Mv}; {G0-eye(P,1)*eye(1,P)}; {D}}, sol1_std, tol, 'verb', 0))/T;
sol1_mean = dot(tt_ones([nv,1,nt]), amen_mv({{dt*Mv}; {1}; {D}}, sol1(1:nv,1,:), tol, 'verb', 0))/T
sol1_std = sqrt(sol1_std)

% Moments of the conrol
sol2_std = round(sol2, tol);
sol2_std = dot(sol2_std, amen_mv({{dt*Mv}; {G0-eye(P,1)*eye(1,P)}; {D}}, sol2_std, tol, 'verb', 0))/T;
sol2_mean = dot(tt_ones([nv,1,nt]), amen_mv({{dt*Mv}; {1}; {D}}, sol2(:,1,:), tol, 'verb', 0))/T
sol2_std = sqrt(sol2_std)

% % Save the solution for visualization purposes
% save(sprintf('3dsol_%s.mat', expname), 'sol1','sol2','sol3');
